@@ -44,8 +44,6 @@ type File struct {
 	qIno openfiletable.QIno
 	// Entry in the open file table
 	fileTableEntry *openfiletable.Entry
-	// go-fuse nodefs.loopbackFile
-	loopbackFile nodefs.File
 	// Store where the last byte was written
 	lastWrittenOffset int64
 	// The opCount is used to judge whether "lastWrittenOffset" is still
@@ -75,14 +73,12 @@ func NewFile(fd *os.File, fs *FS) (*File, fuse.Status) {
 		contentEnc:     fs.contentEnc,
 		qIno:           qi,
 		fileTableEntry: e,
-		loopbackFile:   nodefs.NewLoopbackFile(fd),
 		fs:             fs,
 		File:           nodefs.NewDefaultFile(),
 	}, fuse.OK
 }
 
-// intFd - return the backing file descriptor as an integer. Used for debug
-// messages.
+// intFd - return the backing file descriptor as an integer.
 func (f *File) intFd() int {
 	return int(f.fd.Fd())
 }
@@ -120,7 +116,7 @@ func (f *File) createHeader() (fileID []byte, err error) {
 	buf := h.Pack()
 	// Prevent partially written (=corrupt) header by preallocating the space beforehand
 	if !f.fs.args.NoPrealloc {
-		err = syscallcompat.EnospcPrealloc(int(f.fd.Fd()), 0, contentenc.HeaderLen)
+		err = syscallcompat.EnospcPrealloc(f.intFd(), 0, contentenc.HeaderLen)
 		if err != nil {
 			if !syscallcompat.IsENOSPC(err) {
 				tlog.Warn.Printf("ino%d: createHeader: prealloc failed: %s\n", f.qIno.Ino, err.Error())
@@ -162,10 +158,8 @@ func (f *File) doRead(dst []byte, off uint64, length uint64) ([]byte, fuse.Statu
 				// Empty file
 				return nil, fuse.OK
 			}
-			if err != nil {
-				tlog.Warn.Printf("doRead %d: corrupt header: %v", f.qIno.Ino, err)
-				return nil, fuse.EIO
-			}
+			tlog.Warn.Printf("doRead %d: corrupt header: %v", f.qIno.Ino, err)
+			return nil, fuse.EIO
 		}
 		// Save into the file table
 		f.fileTableEntry.ID = fileID
@@ -245,7 +239,7 @@ func (f *File) Read(buf []byte, off int64) (resultData fuse.ReadResult, code fus
 	f.fileTableEntry.ContentLock.RLock()
 	defer f.fileTableEntry.ContentLock.RUnlock()
 
-	tlog.Debug.Printf("ino%d: FUSE Read: offset=%d length=%d", f.qIno.Ino, len(buf), off)
+	tlog.Debug.Printf("ino%d: FUSE Read: offset=%d length=%d", f.qIno.Ino, off, len(buf))
 	if f.fs.args.SerializeReads {
 		serialize_reads.Wait(off, len(buf))
 	}
@@ -310,7 +304,7 @@ func (f *File) doWrite(data []byte, off int64) (uint32, fuse.Status) {
 			tlog.Debug.Printf("len(oldData)=%d len(blockData)=%d", len(oldData), len(blockData))
 		}
 		tlog.Debug.Printf("ino%d: Writing %d bytes to block #%d",
-			f.qIno.Ino, uint64(len(blockData))-f.contentEnc.BlockOverhead(), b.BlockNo)
+			f.qIno.Ino, len(blockData), b.BlockNo)
 		// Write into the to-encrypt list
 		toEncrypt[i] = blockData
 	}
@@ -321,7 +315,7 @@ func (f *File) doWrite(data []byte, off int64) (uint32, fuse.Status) {
 	var err error
 	cOff := int64(blocks[0].BlockCipherOff())
 	if !f.fs.args.NoPrealloc {
-		err = syscallcompat.EnospcPrealloc(int(f.fd.Fd()), cOff, int64(len(ciphertext)))
+		err = syscallcompat.EnospcPrealloc(f.intFd(), cOff, int64(len(ciphertext)))
 		if err != nil {
 			if !syscallcompat.IsENOSPC(err) {
 				tlog.Warn.Printf("ino%d fh%d: doWrite: prealloc failed: %v", f.qIno.Ino, f.intFd(), err)
@@ -329,7 +323,7 @@ func (f *File) doWrite(data []byte, off int64) (uint32, fuse.Status) {
 			if fileWasEmpty {
 				// Kill the file header again
 				f.fileTableEntry.ID = nil
-				err2 := syscall.Ftruncate(int(f.fd.Fd()), 0)
+				err2 := syscall.Ftruncate(f.intFd(), 0)
 				if err2 != nil {
 					tlog.Warn.Printf("ino%d fh%d: doWrite: rollback failed: %v", f.qIno.Ino, f.intFd(), err2)
 				}
@@ -342,7 +336,8 @@ func (f *File) doWrite(data []byte, off int64) (uint32, fuse.Status) {
 	// Return memory to CReqPool
 	f.fs.contentEnc.CReqPool.Put(ciphertext)
 	if err != nil {
-		tlog.Warn.Printf("doWrite: Write failed: %s", err.Error())
+		tlog.Warn.Printf("ino%d fh%d: doWrite: WriteAt off=%d len=%d failed: %v",
+			f.qIno.Ino, f.intFd(), cOff, len(ciphertext), err)
 		return 0, fuse.ToStatus(err)
 	}
 	return uint32(len(data)), fuse.OK
@@ -415,7 +410,7 @@ func (f *File) Flush() fuse.Status {
 	// Since Flush() may be called for each dup'd fd, we don't
 	// want to really close the file, we just want to flush. This
 	// is achieved by closing a dup'd fd.
-	newFd, err := syscall.Dup(int(f.fd.Fd()))
+	newFd, err := syscall.Dup(f.intFd())
 
 	if err != nil {
 		return fuse.ToStatus(err)
@@ -429,7 +424,7 @@ func (f *File) Fsync(flags int) (code fuse.Status) {
 	f.fdLock.RLock()
 	defer f.fdLock.RUnlock()
 
-	return fuse.ToStatus(syscall.Fsync(int(f.fd.Fd())))
+	return fuse.ToStatus(syscall.Fsync(f.intFd()))
 }
 
 // Chmod FUSE call
@@ -458,7 +453,7 @@ func (f *File) GetAttr(a *fuse.Attr) fuse.Status {
 
 	tlog.Debug.Printf("file.GetAttr()")
 	st := syscall.Stat_t{}
-	err := syscall.Fstat(int(f.fd.Fd()), &st)
+	err := syscall.Fstat(f.intFd(), &st)
 	if err != nil {
 		return fuse.ToStatus(err)
 	}
@@ -475,5 +470,6 @@ func (f *File) GetAttr(a *fuse.Attr) fuse.Status {
 func (f *File) Utimens(a *time.Time, m *time.Time) fuse.Status {
 	f.fdLock.RLock()
 	defer f.fdLock.RUnlock()
-	return f.loopbackFile.Utimens(a, m)
+	err := syscallcompat.FutimesNano(f.intFd(), a, m)
+	return fuse.ToStatus(err)
 }
