@@ -5,7 +5,6 @@ package fusefrontend
 
 import (
 	"os"
-	"path/filepath"
 	"sync"
 	"syscall"
 	"time"
@@ -16,6 +15,7 @@ import (
 	"github.com/hanwen/go-fuse/fuse/nodefs"
 	"github.com/hanwen/go-fuse/fuse/pathfs"
 
+	"github.com/rfjakob/gocryptfs/internal/configfile"
 	"github.com/rfjakob/gocryptfs/internal/contentenc"
 	"github.com/rfjakob/gocryptfs/internal/nametransform"
 	"github.com/rfjakob/gocryptfs/internal/serialize_reads"
@@ -25,8 +25,11 @@ import (
 
 // FS implements the go-fuse virtual filesystem interface.
 type FS struct {
-	pathfs.FileSystem      // loopbackFileSystem, see go-fuse/fuse/pathfs/loopback.go
-	args              Args // Stores configuration arguments
+	// Embed pathfs.defaultFileSystem to avoid compile failure when the
+	// pathfs.FileSystem interface gets new functions. defaultFileSystem
+	// provides a no-op implementation for all functions.
+	pathfs.FileSystem
+	args Args // Stores configuration arguments
 	// dirIVLock: Lock()ed if any "gocryptfs.diriv" file is modified
 	// Readers must RLock() it to prevent them from seeing intermediate
 	// states
@@ -47,9 +50,16 @@ type FS struct {
 	// "gocryptfs -fsck" reads from the channel to also catch these transparently-
 	// mitigated corruptions.
 	MitigatedCorruptions chan string
+	// Track accesses to the filesystem so that we can know when to autounmount.
+	// An access is considered to have happened on every call to encryptPath,
+	// which is called as part of every filesystem operation.
+	// (This flag uses a uint32 so that it can be reset with CompareAndSwapUint32.)
+	AccessedSinceLastCheck uint32
+
+	dirCache dirCacheStruct
 }
 
-var _ pathfs.FileSystem = &FS{} // Verify that interface is implemented.
+//var _ pathfs.FileSystem = &FS{} // Verify that interface is implemented.
 
 // NewFS returns a new encrypted FUSE overlay filesystem.
 func NewFS(args Args, c *contentenc.ContentEnc, n *nametransform.NameTransform) *FS {
@@ -60,7 +70,7 @@ func NewFS(args Args, c *contentenc.ContentEnc, n *nametransform.NameTransform)
 		tlog.Warn.Printf("Forward mode does not support -exclude")
 	}
 	return &FS{
-		FileSystem:    pathfs.NewLoopbackFileSystem(args.Cipherdir),
+		FileSystem:    pathfs.NewDefaultFileSystem(),
 		args:          args,
 		nameTransform: n,
 		contentEnc:    c,
@@ -68,38 +78,45 @@ func NewFS(args Args, c *contentenc.ContentEnc, n *nametransform.NameTransform)
 }
 
 // GetAttr implements pathfs.Filesystem.
-func (fs *FS) GetAttr(name string, context *fuse.Context) (*fuse.Attr, fuse.Status) {
-	tlog.Debug.Printf("FS.GetAttr('%s')", name)
-	if fs.isFiltered(name) {
+//
+// GetAttr is symlink-safe through use of openBackingDir() and Fstatat().
+func (fs *FS) GetAttr(relPath string, context *fuse.Context) (*fuse.Attr, fuse.Status) {
+	tlog.Debug.Printf("FS.GetAttr(%q)", relPath)
+	if fs.isFiltered(relPath) {
 		return nil, fuse.EPERM
 	}
-	cName, err := fs.encryptPath(name)
+	dirfd, cName, err := fs.openBackingDir(relPath)
 	if err != nil {
 		return nil, fuse.ToStatus(err)
 	}
-	a, status := fs.FileSystem.GetAttr(cName, context)
-	if a == nil {
-		tlog.Debug.Printf("FS.GetAttr failed: %s", status.String())
-		return a, status
+	var st unix.Stat_t
+	err = syscallcompat.Fstatat(dirfd, cName, &st, unix.AT_SYMLINK_NOFOLLOW)
+	syscall.Close(dirfd)
+	if err != nil {
+		return nil, fuse.ToStatus(err)
 	}
+	a := &fuse.Attr{}
+	st2 := syscallcompat.Unix2syscall(st)
+	a.FromStat(&st2)
 	if a.IsRegular() {
 		a.Size = fs.contentEnc.CipherSizeToPlainSize(a.Size)
 	} else if a.IsSymlink() {
-		target, _ := fs.Readlink(name, context)
+		target, _ := fs.Readlink(relPath, context)
 		a.Size = uint64(len(target))
 	}
 	if fs.args.ForceOwner != nil {
 		a.Owner = *fs.args.ForceOwner
 	}
-	return a, status
+	return a, fuse.OK
 }
 
 // mangleOpenFlags is used by Create() and Open() to convert the open flags the user
 // wants to the flags we internally use to open the backing file.
+// The returned flags always contain O_NOFOLLOW.
 func (fs *FS) mangleOpenFlags(flags uint32) (newFlags int) {
 	newFlags = int(flags)
 	// Convert WRONLY to RDWR. We always need read access to do read-modify-write cycles.
-	if newFlags&os.O_WRONLY > 0 {
+	if (newFlags & syscall.O_ACCMODE) == syscall.O_WRONLY {
 		newFlags = newFlags ^ os.O_WRONLY | os.O_RDWR
 	}
 	// We also cannot open the file in append mode, we need to seek back for RMW
@@ -109,57 +126,76 @@ func (fs *FS) mangleOpenFlags(flags uint32) (newFlags int) {
 	// accesses. Running xfstests generic/013 on ext4 used to trigger lots of
 	// EINVAL errors due to missing alignment. Just fall back to buffered IO.
 	newFlags = newFlags &^ syscallcompat.O_DIRECT
-
+	// Create and Open are two separate FUSE operations, so O_CREAT should not
+	// be part of the open flags.
+	newFlags = newFlags &^ syscall.O_CREAT
+	// We always want O_NOFOLLOW to be safe against symlink races
+	newFlags |= syscall.O_NOFOLLOW
 	return newFlags
 }
 
-// Open implements pathfs.Filesystem.
+// Open - FUSE call. Open already-existing file.
+//
+// Symlink-safe through Openat().
 func (fs *FS) Open(path string, flags uint32, context *fuse.Context) (fuseFile nodefs.File, status fuse.Status) {
 	if fs.isFiltered(path) {
 		return nil, fuse.EPERM
 	}
+	newFlags := fs.mangleOpenFlags(flags)
 	// Taking this lock makes sure we don't race openWriteOnlyFile()
 	fs.openWriteOnlyLock.RLock()
 	defer fs.openWriteOnlyLock.RUnlock()
-
-	newFlags := fs.mangleOpenFlags(flags)
-	cPath, err := fs.getBackingPath(path)
+	// Symlink-safe open
+	dirfd, cName, err := fs.openBackingDir(path)
 	if err != nil {
-		tlog.Debug.Printf("Open: getBackingPath: %v", err)
 		return nil, fuse.ToStatus(err)
 	}
-	tlog.Debug.Printf("Open: %s", cPath)
-	f, err := os.OpenFile(cPath, newFlags, 0)
+	defer syscall.Close(dirfd)
+	fd, err := syscallcompat.Openat(dirfd, cName, newFlags, 0)
+	// Handle a few specific errors
 	if err != nil {
-		sysErr := err.(*os.PathError).Err
-		if sysErr == syscall.EMFILE {
+		if err == syscall.EMFILE {
 			var lim syscall.Rlimit
 			syscall.Getrlimit(syscall.RLIMIT_NOFILE, &lim)
-			tlog.Warn.Printf("Open %q: too many open files. Current \"ulimit -n\": %d", cPath, lim.Cur)
+			tlog.Warn.Printf("Open %q: too many open files. Current \"ulimit -n\": %d", cName, lim.Cur)
 		}
-		if sysErr == syscall.EACCES && (int(flags)&os.O_WRONLY > 0) {
-			return fs.openWriteOnlyFile(cPath, newFlags)
+		if err == syscall.EACCES && (int(flags)&syscall.O_ACCMODE) == syscall.O_WRONLY {
+			return fs.openWriteOnlyFile(dirfd, cName, newFlags)
 		}
 		return nil, fuse.ToStatus(err)
 	}
+	f := os.NewFile(uintptr(fd), cName)
 	return NewFile(f, fs)
 }
 
+// openBackingFile opens the ciphertext file that backs relative plaintext
+// path "relPath". Always adds O_NOFOLLOW to the flags.
+func (fs *FS) openBackingFile(relPath string, flags int) (fd int, err error) {
+	dirfd, cName, err := fs.openBackingDir(relPath)
+	if err != nil {
+		return -1, err
+	}
+	defer syscall.Close(dirfd)
+	return syscallcompat.Openat(dirfd, cName, flags|syscall.O_NOFOLLOW, 0)
+}
+
 // Due to RMW, we always need read permissions on the backing file. This is a
 // problem if the file permissions do not allow reading (i.e. 0200 permissions).
 // This function works around that problem by chmod'ing the file, obtaining a fd,
 // and chmod'ing it back.
-func (fs *FS) openWriteOnlyFile(cPath string, newFlags int) (fuseFile nodefs.File, status fuse.Status) {
-	woFd, err := os.OpenFile(cPath, os.O_WRONLY, 0)
+func (fs *FS) openWriteOnlyFile(dirfd int, cName string, newFlags int) (*File, fuse.Status) {
+	woFd, err := syscallcompat.Openat(dirfd, cName, syscall.O_WRONLY|syscall.O_NOFOLLOW, 0)
 	if err != nil {
 		return nil, fuse.ToStatus(err)
 	}
-	defer woFd.Close()
-	fi, err := woFd.Stat()
+	defer syscall.Close(woFd)
+	var st syscall.Stat_t
+	err = syscall.Fstat(woFd, &st)
 	if err != nil {
 		return nil, fuse.ToStatus(err)
 	}
-	perms := fi.Mode().Perm()
+	// The cast to uint32 fixes a build failure on Darwin, where st.Mode is uint16.
+	perms := uint32(st.Mode)
 	// Verify that we don't have read permissions
 	if perms&0400 != 0 {
 		tlog.Warn.Printf("openWriteOnlyFile: unexpected permissions %#o, returning EPERM", perms)
@@ -173,162 +209,147 @@ func (fs *FS) openWriteOnlyFile(cPath string, newFlags int) (fuseFile nodefs.Fil
 		fs.openWriteOnlyLock.RLock()
 	}()
 	// Relax permissions and revert on return
-	err = woFd.Chmod(perms | 0400)
+	err = syscall.Fchmod(woFd, perms|0400)
 	if err != nil {
 		tlog.Warn.Printf("openWriteOnlyFile: changing permissions failed: %v", err)
 		return nil, fuse.ToStatus(err)
 	}
 	defer func() {
-		err2 := woFd.Chmod(perms)
+		err2 := syscall.Fchmod(woFd, perms)
 		if err2 != nil {
 			tlog.Warn.Printf("openWriteOnlyFile: reverting permissions failed: %v", err2)
 		}
 	}()
-	rwFd, err := os.OpenFile(cPath, newFlags, 0)
+	rwFd, err := syscallcompat.Openat(dirfd, cName, newFlags, 0)
 	if err != nil {
 		return nil, fuse.ToStatus(err)
 	}
-	return NewFile(rwFd, fs)
+	f := os.NewFile(uintptr(rwFd), cName)
+	return NewFile(f, fs)
 }
 
-// Create implements pathfs.Filesystem.
-func (fs *FS) Create(path string, flags uint32, mode uint32, context *fuse.Context) (fuseFile nodefs.File, code fuse.Status) {
+// Create - FUSE call. Creates a new file.
+//
+// Symlink-safe through the use of Openat().
+func (fs *FS) Create(path string, flags uint32, mode uint32, context *fuse.Context) (nodefs.File, fuse.Status) {
 	if fs.isFiltered(path) {
 		return nil, fuse.EPERM
 	}
 	newFlags := fs.mangleOpenFlags(flags)
-	cPath, err := fs.getBackingPath(path)
+	dirfd, cName, err := fs.openBackingDir(path)
 	if err != nil {
 		return nil, fuse.ToStatus(err)
 	}
-
-	var fd *os.File
-	cName := filepath.Base(cPath)
-
+	defer syscall.Close(dirfd)
+	fd := -1
+	// Make sure context is nil if we don't want to preserve the owner
+	if !fs.args.PreserveOwner {
+		context = nil
+	}
 	// Handle long file name
 	if !fs.args.PlaintextNames && nametransform.IsLongContent(cName) {
-		var dirfd *os.File
-		dirfd, err = os.Open(filepath.Dir(cPath))
-		if err != nil {
-			return nil, fuse.ToStatus(err)
-		}
-		defer dirfd.Close()
-
 		// Create ".name"
-		err = fs.nameTransform.WriteLongName(dirfd, cName, path)
+		err = fs.nameTransform.WriteLongNameAt(dirfd, cName, path)
 		if err != nil {
 			return nil, fuse.ToStatus(err)
 		}
-
 		// Create content
-		var fdRaw int
-		fdRaw, err = syscallcompat.Openat(int(dirfd.Fd()), cName, newFlags|os.O_CREATE|os.O_EXCL, mode)
+		fd, err = syscallcompat.OpenatUser(dirfd, cName, newFlags|syscall.O_CREAT|syscall.O_EXCL, mode, context)
 		if err != nil {
-			nametransform.DeleteLongName(dirfd, cName)
-			return nil, fuse.ToStatus(err)
+			nametransform.DeleteLongNameAt(dirfd, cName)
 		}
-		fd = os.NewFile(uintptr(fdRaw), cName)
 	} else {
-		// Normal (short) file name
-		fd, err = os.OpenFile(cPath, newFlags|os.O_CREATE|os.O_EXCL, os.FileMode(mode))
-		if err != nil {
-			return nil, fuse.ToStatus(err)
-		}
+		// Create content, normal (short) file name
+		fd, err = syscallcompat.OpenatUser(dirfd, cName, newFlags|syscall.O_CREAT|syscall.O_EXCL, mode, context)
 	}
-	// Set owner
-	if fs.args.PreserveOwner {
-		err = fd.Chown(int(context.Owner.Uid), int(context.Owner.Gid))
-		if err != nil {
-			tlog.Warn.Printf("Create: fd.Chown failed: %v", err)
+	if err != nil {
+		// xfstests generic/488 triggers this
+		if err == syscall.EMFILE {
+			var lim syscall.Rlimit
+			syscall.Getrlimit(syscall.RLIMIT_NOFILE, &lim)
+			tlog.Warn.Printf("Create %q: too many open files. Current \"ulimit -n\": %d", cName, lim.Cur)
 		}
+		return nil, fuse.ToStatus(err)
 	}
-	return NewFile(fd, fs)
+	f := os.NewFile(uintptr(fd), cName)
+	return NewFile(f, fs)
 }
 
-// Chmod implements pathfs.Filesystem.
+// Chmod - FUSE call. Change permissions on "path".
+//
+// Symlink-safe through use of Fchmodat().
 func (fs *FS) Chmod(path string, mode uint32, context *fuse.Context) (code fuse.Status) {
 	if fs.isFiltered(path) {
 		return fuse.EPERM
 	}
-	dirfd, cName, err := fs.openBackingPath(path)
+	dirfd, cName, err := fs.openBackingDir(path)
 	if err != nil {
 		return fuse.ToStatus(err)
 	}
-	defer dirfd.Close()
+	defer syscall.Close(dirfd)
 	// os.Chmod goes through the "syscallMode" translation function that messes
 	// up the suid and sgid bits. So use a syscall directly.
-	err = syscallcompat.Fchmodat(int(dirfd.Fd()), cName, mode, unix.AT_SYMLINK_NOFOLLOW)
+	err = syscallcompat.FchmodatNofollow(dirfd, cName, mode)
 	return fuse.ToStatus(err)
 }
 
-// Chown implements pathfs.Filesystem.
+// Chown - FUSE call. Change the owner of "path".
+//
+// Symlink-safe through use of Fchownat().
 func (fs *FS) Chown(path string, uid uint32, gid uint32, context *fuse.Context) (code fuse.Status) {
 	if fs.isFiltered(path) {
 		return fuse.EPERM
 	}
-	dirfd, cName, err := fs.openBackingPath(path)
+	dirfd, cName, err := fs.openBackingDir(path)
 	if err != nil {
 		return fuse.ToStatus(err)
 	}
-	defer dirfd.Close()
-	code = fuse.ToStatus(syscallcompat.Fchownat(int(dirfd.Fd()), cName, int(uid), int(gid), unix.AT_SYMLINK_NOFOLLOW))
-	if !code.Ok() {
-		return code
-	}
-	if !fs.args.PlaintextNames {
-		// When filename encryption is active, every directory contains
-		// a "gocryptfs.diriv" file. This file should also change the owner.
-		// Instead of checking if "cName" is a directory, we just blindly
-		// execute the chown on "cName/gocryptfs.diriv" and ignore errors.
-		dirIVPath := filepath.Join(cName, nametransform.DirIVFilename)
-		syscallcompat.Fchownat(int(dirfd.Fd()), dirIVPath, int(uid), int(gid), unix.AT_SYMLINK_NOFOLLOW)
-	}
-	return fuse.OK
+	defer syscall.Close(dirfd)
+	err = syscallcompat.Fchownat(dirfd, cName, int(uid), int(gid), unix.AT_SYMLINK_NOFOLLOW)
+	return fuse.ToStatus(err)
 }
 
-// Mknod implements pathfs.Filesystem.
+// Mknod - FUSE call. Create a device file.
+//
+// Symlink-safe through use of Mknodat().
 func (fs *FS) Mknod(path string, mode uint32, dev uint32, context *fuse.Context) (code fuse.Status) {
 	if fs.isFiltered(path) {
 		return fuse.EPERM
 	}
-	dirfd, cName, err := fs.openBackingPath(path)
+	dirfd, cName, err := fs.openBackingDir(path)
 	if err != nil {
 		return fuse.ToStatus(err)
 	}
-	defer dirfd.Close()
+	defer syscall.Close(dirfd)
+	// Make sure context is nil if we don't want to preserve the owner
+	if !fs.args.PreserveOwner {
+		context = nil
+	}
 	// Create ".name" file to store long file name (except in PlaintextNames mode)
 	if !fs.args.PlaintextNames && nametransform.IsLongContent(cName) {
-		err = fs.nameTransform.WriteLongName(dirfd, cName, path)
+		err = fs.nameTransform.WriteLongNameAt(dirfd, cName, path)
 		if err != nil {
 			return fuse.ToStatus(err)
 		}
 		// Create "gocryptfs.longfile." device node
-		err = syscallcompat.Mknodat(int(dirfd.Fd()), cName, mode, int(dev))
+		err = syscallcompat.MknodatUser(dirfd, cName, mode, int(dev), context)
 		if err != nil {
-			nametransform.DeleteLongName(dirfd, cName)
+			nametransform.DeleteLongNameAt(dirfd, cName)
 		}
 	} else {
 		// Create regular device node
-		err = syscallcompat.Mknodat(int(dirfd.Fd()), cName, mode, int(dev))
+		err = syscallcompat.MknodatUser(dirfd, cName, mode, int(dev), context)
 	}
-	if err != nil {
-		return fuse.ToStatus(err)
-	}
-	// Set owner
-	if fs.args.PreserveOwner {
-		err = syscallcompat.Fchownat(int(dirfd.Fd()), cName, int(context.Owner.Uid),
-			int(context.Owner.Gid), unix.AT_SYMLINK_NOFOLLOW)
-		if err != nil {
-			tlog.Warn.Printf("Mknod: Fchownat failed: %v", err)
-		}
-	}
-	return fuse.OK
+	return fuse.ToStatus(err)
 }
 
-// Truncate implements pathfs.Filesystem.
+// Truncate - FUSE call. Truncates a file.
+//
 // Support truncate(2) by opening the file and calling ftruncate(2)
 // While the glibc "truncate" wrapper seems to always use ftruncate, fsstress from
 // xfstests uses this a lot by calling "truncate64" directly.
+//
+// Symlink-safe by letting file.Truncate() do all the work.
 func (fs *FS) Truncate(path string, offset uint64, context *fuse.Context) (code fuse.Status) {
 	file, code := fs.Open(path, uint32(os.O_RDWR), context)
 	if code != fuse.OK {
@@ -339,33 +360,41 @@ func (fs *FS) Truncate(path string, offset uint64, context *fuse.Context) (code
 	return code
 }
 
-// Utimens implements pathfs.Filesystem.
+// Utimens - FUSE call. Set the timestamps on file "path".
+//
+// Symlink-safe through UtimesNanoAt.
 func (fs *FS) Utimens(path string, a *time.Time, m *time.Time, context *fuse.Context) (code fuse.Status) {
 	if fs.isFiltered(path) {
 		return fuse.EPERM
 	}
-	cPath, err := fs.encryptPath(path)
+	dirfd, cName, err := fs.openBackingDir(path)
 	if err != nil {
 		return fuse.ToStatus(err)
 	}
-	return fs.FileSystem.Utimens(cPath, a, m, context)
+	defer syscall.Close(dirfd)
+	err = syscallcompat.UtimesNanoAtNofollow(dirfd, cName, a, m)
+	return fuse.ToStatus(err)
 }
 
-// StatFs implements pathfs.Filesystem.
+// StatFs - FUSE call. Returns information about the filesystem.
+//
+// Symlink-safe because the passed path is ignored.
 func (fs *FS) StatFs(path string) *fuse.StatfsOut {
-	if fs.isFiltered(path) {
-		return nil
+	var st syscall.Statfs_t
+	err := syscall.Statfs(fs.args.Cipherdir, &st)
+	if err == nil {
+		var out fuse.StatfsOut
+		out.FromStatfsT(&st)
+		return &out
 	}
-	cPath, err := fs.encryptPath(path)
-	if err != nil {
-		return nil
-	}
-	return fs.FileSystem.StatFs(cPath)
+	return nil
 }
 
 // decryptSymlinkTarget: "cData64" is base64-decoded and decrypted
 // like file contents (GCM).
 // The empty string decrypts to the empty string.
+//
+// This function does not do any I/O and is hence symlink-safe.
 func (fs *FS) decryptSymlinkTarget(cData64 string) (string, error) {
 	if cData64 == "" {
 		return "", nil
@@ -381,14 +410,16 @@ func (fs *FS) decryptSymlinkTarget(cData64 string) (string, error) {
 	return string(data), nil
 }
 
-// Readlink implements pathfs.Filesystem.
+// Readlink - FUSE call.
+//
+// Symlink-safe through openBackingDir() + Readlinkat().
 func (fs *FS) Readlink(relPath string, context *fuse.Context) (out string, status fuse.Status) {
-	cPath, err := fs.encryptPath(relPath)
+	dirfd, cName, err := fs.openBackingDir(relPath)
 	if err != nil {
 		return "", fuse.ToStatus(err)
 	}
-	cAbsPath := filepath.Join(fs.args.Cipherdir, cPath)
-	cTarget, err := os.Readlink(cAbsPath)
+	defer syscall.Close(dirfd)
+	cTarget, err := syscallcompat.Readlinkat(dirfd, cName)
 	if err != nil {
 		return "", fuse.ToStatus(err)
 	}
@@ -398,30 +429,32 @@ func (fs *FS) Readlink(relPath string, context *fuse.Context) (out string, statu
 	// Symlinks are encrypted like file contents (GCM) and base64-encoded
 	target, err := fs.decryptSymlinkTarget(cTarget)
 	if err != nil {
-		tlog.Warn.Printf("Readlink %q: decrypting target failed: %v", cPath, err)
+		tlog.Warn.Printf("Readlink %q: decrypting target failed: %v", cName, err)
 		return "", fuse.EIO
 	}
 	return string(target), fuse.OK
 }
 
-// Unlink implements pathfs.Filesystem.
+// Unlink - FUSE call. Delete a file.
+//
+// Symlink-safe through use of Unlinkat().
 func (fs *FS) Unlink(path string, context *fuse.Context) (code fuse.Status) {
 	if fs.isFiltered(path) {
 		return fuse.EPERM
 	}
-	dirfd, cName, err := fs.openBackingPath(path)
+	dirfd, cName, err := fs.openBackingDir(path)
 	if err != nil {
 		return fuse.ToStatus(err)
 	}
-	defer dirfd.Close()
+	defer syscall.Close(dirfd)
 	// Delete content
-	err = syscallcompat.Unlinkat(int(dirfd.Fd()), cName, 0)
+	err = syscallcompat.Unlinkat(dirfd, cName, 0)
 	if err != nil {
 		return fuse.ToStatus(err)
 	}
 	// Delete ".name" file
 	if !fs.args.PlaintextNames && nametransform.IsLongContent(cName) {
-		err = nametransform.DeleteLongName(dirfd, cName)
+		err = nametransform.DeleteLongNameAt(dirfd, cName)
 		if err != nil {
 			tlog.Warn.Printf("Unlink: could not delete .name file: %v", err)
 		}
@@ -432,6 +465,8 @@ func (fs *FS) Unlink(path string, context *fuse.Context) (code fuse.Status) {
 // encryptSymlinkTarget: "data" is encrypted like file contents (GCM)
 // and base64-encoded.
 // The empty string encrypts to the empty string.
+//
+// Symlink-safe because it does not do any I/O.
 func (fs *FS) encryptSymlinkTarget(data string) (cData64 string) {
 	if data == "" {
 		return ""
@@ -441,17 +476,23 @@ func (fs *FS) encryptSymlinkTarget(data string) (cData64 string) {
 	return cData64
 }
 
-// Symlink implements pathfs.Filesystem.
+// Symlink - FUSE call. Create a symlink.
+//
+// Symlink-safe through use of Symlinkat.
 func (fs *FS) Symlink(target string, linkName string, context *fuse.Context) (code fuse.Status) {
 	tlog.Debug.Printf("Symlink(\"%s\", \"%s\")", target, linkName)
 	if fs.isFiltered(linkName) {
 		return fuse.EPERM
 	}
-	dirfd, cName, err := fs.openBackingPath(linkName)
+	dirfd, cName, err := fs.openBackingDir(linkName)
 	if err != nil {
 		return fuse.ToStatus(err)
 	}
-	defer dirfd.Close()
+	defer syscall.Close(dirfd)
+	// Make sure context is nil if we don't want to preserve the owner
+	if !fs.args.PreserveOwner {
+		context = nil
+	}
 	cTarget := target
 	if !fs.args.PlaintextNames {
 		// Symlinks are encrypted like file contents (GCM) and base64-encoded
@@ -459,97 +500,60 @@ func (fs *FS) Symlink(target string, linkName string, context *fuse.Context) (co
 	}
 	// Create ".name" file to store long file name (except in PlaintextNames mode)
 	if !fs.args.PlaintextNames && nametransform.IsLongContent(cName) {
-		err = fs.nameTransform.WriteLongName(dirfd, cName, linkName)
+		err = fs.nameTransform.WriteLongNameAt(dirfd, cName, linkName)
 		if err != nil {
 			return fuse.ToStatus(err)
 		}
 		// Create "gocryptfs.longfile." symlink
-		err = syscallcompat.Symlinkat(cTarget, int(dirfd.Fd()), cName)
+		err = syscallcompat.SymlinkatUser(cTarget, dirfd, cName, context)
 		if err != nil {
-			nametransform.DeleteLongName(dirfd, cName)
+			nametransform.DeleteLongNameAt(dirfd, cName)
 		}
 	} else {
 		// Create symlink
-		err = syscallcompat.Symlinkat(cTarget, int(dirfd.Fd()), cName)
+		err = syscallcompat.SymlinkatUser(cTarget, dirfd, cName, context)
 	}
-	if err != nil {
-		return fuse.ToStatus(err)
-	}
-	// Set owner
-	if fs.args.PreserveOwner {
-		err = syscallcompat.Fchownat(int(dirfd.Fd()), cName, int(context.Owner.Uid),
-			int(context.Owner.Gid), unix.AT_SYMLINK_NOFOLLOW)
-		if err != nil {
-			tlog.Warn.Printf("Symlink: Fchownat failed: %v", err)
-		}
-	}
-	return fuse.OK
+	return fuse.ToStatus(err)
 }
 
-// Rename implements pathfs.Filesystem.
+// Rename - FUSE call.
+//
+// Symlink-safe through Renameat().
 func (fs *FS) Rename(oldPath string, newPath string, context *fuse.Context) (code fuse.Status) {
+	defer fs.dirCache.Clear()
 	if fs.isFiltered(newPath) {
 		return fuse.EPERM
 	}
-	cOldPath, err := fs.getBackingPath(oldPath)
+	oldDirfd, oldCName, err := fs.openBackingDir(oldPath)
 	if err != nil {
 		return fuse.ToStatus(err)
 	}
-	cNewPath, err := fs.getBackingPath(newPath)
+	defer syscall.Close(oldDirfd)
+	newDirfd, newCName, err := fs.openBackingDir(newPath)
 	if err != nil {
 		return fuse.ToStatus(err)
 	}
-	// The Rename may cause a directory to take the place of another directory.
-	// That directory may still be in the DirIV cache, clear it.
-	fs.nameTransform.DirIVCache.Clear()
+	defer syscall.Close(newDirfd)
 	// Easy case.
 	if fs.args.PlaintextNames {
-		return fuse.ToStatus(syscall.Rename(cOldPath, cNewPath))
+		return fuse.ToStatus(syscallcompat.Renameat(oldDirfd, oldCName, newDirfd, newCName))
 	}
-	// Handle long source file name
-	var oldDirFd *os.File
-	var finalOldDirFd int
-	var finalOldPath = cOldPath
-	cOldName := filepath.Base(cOldPath)
-	if nametransform.IsLongContent(cOldName) {
-		oldDirFd, err = os.Open(filepath.Dir(cOldPath))
-		if err != nil {
-			return fuse.ToStatus(err)
-		}
-		defer oldDirFd.Close()
-		finalOldDirFd = int(oldDirFd.Fd())
-		// Use relative path
-		finalOldPath = cOldName
-	}
-	// Handle long destination file name
-	var newDirFd *os.File
-	var finalNewDirFd int
-	var finalNewPath = cNewPath
-	cNewName := filepath.Base(cNewPath)
-	if nametransform.IsLongContent(cNewName) {
-		newDirFd, err = os.Open(filepath.Dir(cNewPath))
-		if err != nil {
-			return fuse.ToStatus(err)
-		}
-		defer newDirFd.Close()
-		finalNewDirFd = int(newDirFd.Fd())
-		// Use relative path
-		finalNewPath = cNewName
-		// Create destination .name file
-		err = fs.nameTransform.WriteLongName(newDirFd, cNewName, newPath)
+	// Long destination file name: create .name file
+	nameFileAlreadyThere := false
+	if nametransform.IsLongContent(newCName) {
+		err = fs.nameTransform.WriteLongNameAt(newDirfd, newCName, newPath)
 		// Failure to write the .name file is expected when the target path already
 		// exists. Since hashes are pretty unique, there is no need to modify the
-		// file anyway. We still set newDirFd to nil to ensure that we do not delete
-		// the file on error.
+		// .name file in this case, and we ignore the error.
 		if err == syscall.EEXIST {
-			newDirFd = nil
+			nameFileAlreadyThere = true
 		} else if err != nil {
 			return fuse.ToStatus(err)
 		}
 	}
 	// Actual rename
-	tlog.Debug.Printf("Renameat oldfd=%d oldpath=%s newfd=%d newpath=%s\n", finalOldDirFd, finalOldPath, finalNewDirFd, finalNewPath)
-	err = syscallcompat.Renameat(finalOldDirFd, finalOldPath, finalNewDirFd, finalNewPath)
+	tlog.Debug.Printf("Renameat %d/%s -> %d/%s\n", oldDirfd, oldCName, newDirfd, newCName)
+	err = syscallcompat.Renameat(oldDirfd, oldCName, newDirfd, newCName)
 	if err == syscall.ENOTEMPTY || err == syscall.EEXIST {
 		// If an empty directory is overwritten we will always get an error as
 		// the "empty" directory will still contain gocryptfs.diriv.
@@ -558,65 +562,83 @@ func (fs *FS) Rename(oldPath string, newPath string, context *fuse.Context) (cod
 		// again.
 		tlog.Debug.Printf("Rename: Handling ENOTEMPTY")
 		if fs.Rmdir(newPath, context) == fuse.OK {
-			err = syscallcompat.Renameat(finalOldDirFd, finalOldPath, finalNewDirFd, finalNewPath)
+			err = syscallcompat.Renameat(oldDirfd, oldCName, newDirfd, newCName)
 		}
 	}
 	if err != nil {
-		if newDirFd != nil {
-			// Roll back .name creation
-			nametransform.DeleteLongName(newDirFd, cNewName)
+		if nametransform.IsLongContent(newCName) && nameFileAlreadyThere == false {
+			// Roll back .name creation unless the .name file was already there
+			nametransform.DeleteLongNameAt(newDirfd, newCName)
 		}
 		return fuse.ToStatus(err)
 	}
-	if oldDirFd != nil {
-		nametransform.DeleteLongName(oldDirFd, cOldName)
+	if nametransform.IsLongContent(oldCName) {
+		nametransform.DeleteLongNameAt(oldDirfd, oldCName)
 	}
 	return fuse.OK
 }
 
-// Link implements pathfs.Filesystem.
+// Link - FUSE call. Creates a hard link at "newPath" pointing to file
+// "oldPath".
+//
+// Symlink-safe through use of Linkat().
 func (fs *FS) Link(oldPath string, newPath string, context *fuse.Context) (code fuse.Status) {
 	if fs.isFiltered(newPath) {
 		return fuse.EPERM
 	}
-	oldDirFd, cOldName, err := fs.openBackingPath(oldPath)
+	oldDirFd, cOldName, err := fs.openBackingDir(oldPath)
 	if err != nil {
 		return fuse.ToStatus(err)
 	}
-	defer oldDirFd.Close()
-	newDirFd, cNewName, err := fs.openBackingPath(newPath)
+	defer syscall.Close(oldDirFd)
+	newDirFd, cNewName, err := fs.openBackingDir(newPath)
 	if err != nil {
 		return fuse.ToStatus(err)
 	}
-	defer newDirFd.Close()
+	defer syscall.Close(newDirFd)
 	// Handle long file name (except in PlaintextNames mode)
 	if !fs.args.PlaintextNames && nametransform.IsLongContent(cNewName) {
-		err = fs.nameTransform.WriteLongName(newDirFd, cNewName, newPath)
+		err = fs.nameTransform.WriteLongNameAt(newDirFd, cNewName, newPath)
 		if err != nil {
 			return fuse.ToStatus(err)
 		}
 		// Create "gocryptfs.longfile." link
-		err = syscallcompat.Linkat(int(oldDirFd.Fd()), cOldName, int(newDirFd.Fd()), cNewName, 0)
+		err = syscallcompat.Linkat(oldDirFd, cOldName, newDirFd, cNewName, 0)
 		if err != nil {
-			nametransform.DeleteLongName(newDirFd, cNewName)
+			nametransform.DeleteLongNameAt(newDirFd, cNewName)
 		}
 	} else {
 		// Create regular link
-		err = syscallcompat.Linkat(int(oldDirFd.Fd()), cOldName, int(newDirFd.Fd()), cNewName, 0)
+		err = syscallcompat.Linkat(oldDirFd, cOldName, newDirFd, cNewName, 0)
 	}
 	return fuse.ToStatus(err)
 }
 
-// Access implements pathfs.Filesystem.
-func (fs *FS) Access(path string, mode uint32, context *fuse.Context) (code fuse.Status) {
-	if fs.isFiltered(path) {
+// Access - FUSE call. Check if a file can be accessed in the specified mode(s)
+// (read, write, execute).
+//
+// From https://github.com/libfuse/libfuse/blob/master/include/fuse.h :
+//
+// > Check file access permissions
+// >
+// > If the 'default_permissions' mount option is given, this method is not
+// > called.
+//
+// We always enable default_permissions when -allow_other is passed, so there
+// is no need for this function to check the uid in fuse.Context.
+//
+// Symlink-safe through use of faccessat.
+func (fs *FS) Access(relPath string, mode uint32, context *fuse.Context) (code fuse.Status) {
+	if fs.isFiltered(relPath) {
 		return fuse.EPERM
 	}
-	cPath, err := fs.getBackingPath(path)
+	dirfd, cName, err := fs.openBackingDir(relPath)
 	if err != nil {
 		return fuse.ToStatus(err)
 	}
-	return fuse.ToStatus(syscall.Access(cPath, mode))
+	err = syscallcompat.Faccessat(dirfd, cName, mode)
+	syscall.Close(dirfd)
+	return fuse.ToStatus(err)
 }
 
 // reportMitigatedCorruption is used to report a corruption that was transparently
@@ -635,3 +657,21 @@ func (fs *FS) reportMitigatedCorruption(item string) {
 		return
 	}
 }
+
+// isFiltered - check if plaintext "path" should be forbidden
+//
+// Prevents name clashes with internal files when file names are not encrypted
+func (fs *FS) isFiltered(path string) bool {
+	if !fs.args.PlaintextNames {
+		return false
+	}
+	// gocryptfs.conf in the root directory is forbidden
+	if path == configfile.ConfDefaultName {
+		tlog.Info.Printf("The name /%s is reserved when -plaintextnames is used\n",
+			configfile.ConfDefaultName)
+		return true
+	}
+	// Note: gocryptfs.diriv is NOT forbidden because diriv and plaintextnames
+	// are exclusive
+	return false
+}
