@@ -5,8 +5,6 @@ package fusefrontend
 import (
 	"fmt"
 	"io"
-	"os"
-	"path/filepath"
 	"runtime"
 	"syscall"
 
@@ -26,22 +24,25 @@ const dsStoreName = ".DS_Store"
 // mkdirWithIv - create a new directory and corresponding diriv file. dirfd
 // should be a handle to the parent directory, cName is the name of the new
 // directory and mode specifies the access permissions to use.
-func (fs *FS) mkdirWithIv(dirfd *os.File, cName string, mode uint32) error {
+func (fs *FS) mkdirWithIv(dirfd int, cName string, mode uint32, context *fuse.Context) error {
 	// Between the creation of the directory and the creation of gocryptfs.diriv
 	// the directory is inconsistent. Take the lock to prevent other readers
 	// from seeing it.
 	fs.dirIVLock.Lock()
-	// The new directory may take the place of an older one that is still in the cache
-	fs.nameTransform.DirIVCache.Clear()
 	defer fs.dirIVLock.Unlock()
-	err := syscallcompat.Mkdirat(int(dirfd.Fd()), cName, mode)
+	err := syscallcompat.MkdiratUser(dirfd, cName, mode, context)
 	if err != nil {
 		return err
 	}
-	// Create gocryptfs.diriv
-	err = nametransform.WriteDirIV(dirfd, cName)
+	dirfd2, err := syscallcompat.Openat(dirfd, cName, syscall.O_DIRECTORY|syscall.O_NOFOLLOW|syscallcompat.O_PATH, 0)
+	if err == nil {
+		// Create gocryptfs.diriv
+		err = nametransform.WriteDirIVAt(dirfd2)
+		syscall.Close(dirfd2)
+	}
 	if err != nil {
-		err2 := syscallcompat.Unlinkat(int(dirfd.Fd()), cName, unix.AT_REMOVEDIR)
+		// Delete inconsistent directory (missing gocryptfs.diriv!)
+		err2 := syscallcompat.Unlinkat(dirfd, cName, unix.AT_REMOVEDIR)
 		if err2 != nil {
 			tlog.Warn.Printf("mkdirWithIv: rollback failed: %v", err2)
 		}
@@ -49,151 +50,156 @@ func (fs *FS) mkdirWithIv(dirfd *os.File, cName string, mode uint32) error {
 	return err
 }
 
-// Mkdir implements pathfs.FileSystem
+// Mkdir - FUSE call. Create a directory at "newPath" with permissions "mode".
+//
+// Symlink-safe through use of Mkdirat().
 func (fs *FS) Mkdir(newPath string, mode uint32, context *fuse.Context) (code fuse.Status) {
+	defer fs.dirCache.Clear()
 	if fs.isFiltered(newPath) {
 		return fuse.EPERM
 	}
-	dirfd, cName, err := fs.openBackingPath(newPath)
+	dirfd, cName, err := fs.openBackingDir(newPath)
 	if err != nil {
 		return fuse.ToStatus(err)
 	}
-	defer dirfd.Close()
+	defer syscall.Close(dirfd)
+	// Make sure context is nil if we don't want to preserve the owner
+	if !fs.args.PreserveOwner {
+		context = nil
+	}
 	if fs.args.PlaintextNames {
-		err = syscallcompat.Mkdirat(int(dirfd.Fd()), cName, mode)
-		// Set owner
-		if fs.args.PreserveOwner {
-			err = syscallcompat.Fchownat(int(dirfd.Fd()), cName, int(context.Owner.Uid),
-				int(context.Owner.Gid), unix.AT_SYMLINK_NOFOLLOW)
-			if err != nil {
-				tlog.Warn.Printf("Mkdir: Fchownat failed: %v", err)
-			}
-		}
+		err = syscallcompat.MkdiratUser(dirfd, cName, mode, context)
 		return fuse.ToStatus(err)
 	}
 
-	// We need write and execute permissions to create gocryptfs.diriv
+	// We need write and execute permissions to create gocryptfs.diriv.
+	// Also, we need read permissions to open the directory (to avoid
+	// race-conditions between getting and setting the mode).
 	origMode := mode
-	mode = mode | 0300
+	mode = mode | 0700
 
 	// Handle long file name
 	if nametransform.IsLongContent(cName) {
 		// Create ".name"
-		err = fs.nameTransform.WriteLongName(dirfd, cName, newPath)
+		err = fs.nameTransform.WriteLongNameAt(dirfd, cName, newPath)
 		if err != nil {
 			return fuse.ToStatus(err)
 		}
 
 		// Create directory
-		err = fs.mkdirWithIv(dirfd, cName, mode)
+		err = fs.mkdirWithIv(dirfd, cName, mode, context)
 		if err != nil {
-			nametransform.DeleteLongName(dirfd, cName)
+			nametransform.DeleteLongNameAt(dirfd, cName)
 			return fuse.ToStatus(err)
 		}
 	} else {
-		err = fs.mkdirWithIv(dirfd, cName, mode)
+		err = fs.mkdirWithIv(dirfd, cName, mode, context)
 		if err != nil {
 			return fuse.ToStatus(err)
 		}
 	}
-	// Set permissions back to what the user wanted
+	// Set mode
 	if origMode != mode {
-		err = syscallcompat.Fchmodat(int(dirfd.Fd()), cName, origMode, unix.AT_SYMLINK_NOFOLLOW)
+		dirfd2, err := syscallcompat.Openat(dirfd, cName,
+			syscall.O_RDONLY|syscall.O_DIRECTORY|syscall.O_NOFOLLOW, 0)
 		if err != nil {
-			tlog.Warn.Printf("Mkdir: Fchmodat failed: %v", err)
+			tlog.Warn.Printf("Mkdir %q: Openat failed: %v", cName, err)
+			return fuse.ToStatus(err)
 		}
-	}
-	// Set owner
-	if fs.args.PreserveOwner {
-		err = syscallcompat.Fchownat(int(dirfd.Fd()), cName, int(context.Owner.Uid),
-			int(context.Owner.Gid), unix.AT_SYMLINK_NOFOLLOW)
+		defer syscall.Close(dirfd2)
+
+		var st syscall.Stat_t
+		err = syscall.Fstat(dirfd2, &st)
 		if err != nil {
-			tlog.Warn.Printf("Mkdir: Fchownat 1 failed: %v", err)
+			tlog.Warn.Printf("Mkdir %q: Fstat failed: %v", cName, err)
+			return fuse.ToStatus(err)
 		}
-		err = syscallcompat.Fchownat(int(dirfd.Fd()), filepath.Join(cName, nametransform.DirIVFilename),
-			int(context.Owner.Uid), int(context.Owner.Gid), unix.AT_SYMLINK_NOFOLLOW)
+
+		// Preserve SGID bit if it was set due to inheritance.
+		origMode = uint32(st.Mode&^0777) | origMode
+		err = syscall.Fchmod(dirfd2, origMode)
 		if err != nil {
-			tlog.Warn.Printf("Mkdir: Fchownat 2 failed: %v", err)
+			tlog.Warn.Printf("Mkdir %q: Fchmod %#o -> %#o failed: %v", cName, mode, origMode, err)
+			return fuse.ToStatus(err)
 		}
 	}
 	return fuse.OK
 }
 
 // haveDsstore return true if one of the entries in "names" is ".DS_Store".
-func haveDsstore(names []string) bool {
-	for _, n := range names {
-		if n == dsStoreName {
+func haveDsstore(entries []fuse.DirEntry) bool {
+	for _, e := range entries {
+		if e.Name == dsStoreName {
 			return true
 		}
 	}
 	return false
 }
 
-// Rmdir implements pathfs.FileSystem
-func (fs *FS) Rmdir(path string, context *fuse.Context) (code fuse.Status) {
-	cPath, err := fs.getBackingPath(path)
+// Rmdir - FUSE call.
+//
+// Symlink-safe through Unlinkat() + AT_REMOVEDIR.
+func (fs *FS) Rmdir(relPath string, context *fuse.Context) (code fuse.Status) {
+	defer fs.dirCache.Clear()
+	parentDirFd, cName, err := fs.openBackingDir(relPath)
 	if err != nil {
 		return fuse.ToStatus(err)
 	}
+	defer syscall.Close(parentDirFd)
 	if fs.args.PlaintextNames {
-		err = syscall.Rmdir(cPath)
+		// Unlinkat with AT_REMOVEDIR is equivalent to Rmdir
+		err = unix.Unlinkat(parentDirFd, cName, unix.AT_REMOVEDIR)
 		return fuse.ToStatus(err)
 	}
-	parentDir := filepath.Dir(cPath)
-	parentDirFd, err := os.Open(parentDir)
-	if err != nil {
-		return fuse.ToStatus(err)
-	}
-	defer parentDirFd.Close()
-
-	cName := filepath.Base(cPath)
-	dirfdRaw, err := syscallcompat.Openat(int(parentDirFd.Fd()), cName,
-		syscall.O_RDONLY|syscall.O_NOFOLLOW, 0)
-	if err == syscall.EACCES {
-		// We need permission to read and modify the directory
-		tlog.Debug.Printf("Rmdir: handling EACCESS")
-		// TODO use syscall.Fstatat once it is available in Go
-		var fi os.FileInfo
-		fi, err = os.Lstat(cPath)
+	// Unless we are running as root, we need read, write and execute permissions
+	// to handle gocryptfs.diriv.
+	permWorkaround := false
+	var origMode uint32
+	if !fs.args.PreserveOwner {
+		var st unix.Stat_t
+		err = syscallcompat.Fstatat(parentDirFd, cName, &st, unix.AT_SYMLINK_NOFOLLOW)
 		if err != nil {
-			tlog.Debug.Printf("Rmdir: Stat: %v", err)
 			return fuse.ToStatus(err)
 		}
-		origMode := fi.Mode()
-		// TODO use syscall.Chmodat once it is available in Go
-		err = os.Chmod(cPath, origMode|0700)
-		if err != nil {
-			tlog.Debug.Printf("Rmdir: Chmod failed: %v", err)
-			return fuse.ToStatus(err)
-		}
-		// Retry open
-		var st syscall.Stat_t
-		syscall.Lstat(cPath, &st)
-		dirfdRaw, err = syscallcompat.Openat(int(parentDirFd.Fd()), cName,
-			syscall.O_RDONLY|syscall.O_NOFOLLOW, 0)
-		// Undo the chmod if removing the directory failed
-		defer func() {
-			if code != fuse.OK {
-				err = os.Chmod(cPath, origMode)
-				if err != nil {
-					tlog.Warn.Printf("Rmdir: Chmod rollback failed: %v", err)
-				}
+		if st.Mode&0700 != 0700 {
+			tlog.Debug.Printf("Rmdir: permWorkaround")
+			permWorkaround = true
+			// This cast is needed on Darwin, where st.Mode is uint16.
+			origMode = uint32(st.Mode)
+			err = syscallcompat.FchmodatNofollow(parentDirFd, cName, origMode|0700)
+			if err != nil {
+				tlog.Debug.Printf("Rmdir: permWorkaround: chmod failed: %v", err)
+				return fuse.ToStatus(err)
 			}
-		}()
+		}
 	}
+	dirfd, err := syscallcompat.Openat(parentDirFd, cName,
+		syscall.O_RDONLY|syscall.O_DIRECTORY|syscall.O_NOFOLLOW, 0)
 	if err != nil {
 		tlog.Debug.Printf("Rmdir: Open: %v", err)
 		return fuse.ToStatus(err)
 	}
-	dirfd := os.NewFile(uintptr(dirfdRaw), cName)
-	defer dirfd.Close()
+	defer syscall.Close(dirfd)
+	// Undo the chmod if removing the directory failed. This must run before
+	// closing dirfd, so defer it after (defer is LIFO).
+	if permWorkaround {
+		defer func() {
+			if code != fuse.OK {
+				err = unix.Fchmod(dirfd, origMode)
+				if err != nil {
+					tlog.Warn.Printf("Rmdir: permWorkaround: rollback failed: %v", err)
+				}
+			}
+		}()
+	}
 retry:
 	// Check directory contents
-	children, err := dirfd.Readdirnames(10)
+	children, err := syscallcompat.Getdents(dirfd)
 	if err == io.EOF {
 		// The directory is empty
-		tlog.Warn.Printf("Rmdir: %q: gocryptfs.diriv is missing", cPath)
-		return fuse.ToStatus(syscall.Rmdir(cPath))
+		tlog.Warn.Printf("Rmdir: %q: gocryptfs.diriv is missing", cName)
+		err = unix.Unlinkat(parentDirFd, cName, unix.AT_REMOVEDIR)
+		return fuse.ToStatus(err)
 	}
 	if err != nil {
 		tlog.Warn.Printf("Rmdir: Readdirnames: %v", err)
@@ -202,13 +208,12 @@ retry:
 	// MacOS sprinkles .DS_Store files everywhere. This is hard to avoid for
 	// users, so handle it transparently here.
 	if runtime.GOOS == "darwin" && len(children) <= 2 && haveDsstore(children) {
-		ds := filepath.Join(cPath, dsStoreName)
-		err = syscall.Unlink(ds)
+		err = unix.Unlinkat(dirfd, dsStoreName, 0)
 		if err != nil {
-			tlog.Warn.Printf("Rmdir: failed to delete blocking file %q: %v", ds, err)
+			tlog.Warn.Printf("Rmdir: failed to delete blocking file %q: %v", dsStoreName, err)
 			return fuse.ToStatus(err)
 		}
-		tlog.Warn.Printf("Rmdir: had to delete blocking file %q", ds)
+		tlog.Warn.Printf("Rmdir: had to delete blocking file %q", dsStoreName)
 		goto retry
 	}
 	// If the directory is not empty besides gocryptfs.diriv, do not even
@@ -223,51 +228,52 @@ retry:
 	// Protect against concurrent readers.
 	fs.dirIVLock.Lock()
 	defer fs.dirIVLock.Unlock()
-	err = syscallcompat.Renameat(int(dirfd.Fd()), nametransform.DirIVFilename,
-		int(parentDirFd.Fd()), tmpName)
+	err = syscallcompat.Renameat(dirfd, nametransform.DirIVFilename,
+		parentDirFd, tmpName)
 	if err != nil {
 		tlog.Warn.Printf("Rmdir: Renaming %s to %s failed: %v",
 			nametransform.DirIVFilename, tmpName, err)
 		return fuse.ToStatus(err)
 	}
 	// Actual Rmdir
-	err = syscallcompat.Unlinkat(int(parentDirFd.Fd()), cName, unix.AT_REMOVEDIR)
+	err = syscallcompat.Unlinkat(parentDirFd, cName, unix.AT_REMOVEDIR)
 	if err != nil {
 		// This can happen if another file in the directory was created in the
 		// meantime, undo the rename
-		err2 := syscallcompat.Renameat(int(parentDirFd.Fd()), tmpName,
-			int(dirfd.Fd()), nametransform.DirIVFilename)
-		if err != nil {
+		err2 := syscallcompat.Renameat(parentDirFd, tmpName,
+			dirfd, nametransform.DirIVFilename)
+		if err2 != nil {
 			tlog.Warn.Printf("Rmdir: Rename rollback failed: %v", err2)
 		}
 		return fuse.ToStatus(err)
 	}
 	// Delete "gocryptfs.diriv.rmdir.XYZ"
-	err = syscallcompat.Unlinkat(int(parentDirFd.Fd()), tmpName, 0)
+	err = syscallcompat.Unlinkat(parentDirFd, tmpName, 0)
 	if err != nil {
 		tlog.Warn.Printf("Rmdir: Could not clean up %s: %v", tmpName, err)
 	}
 	// Delete .name file
 	if nametransform.IsLongContent(cName) {
-		nametransform.DeleteLongName(parentDirFd, cName)
+		nametransform.DeleteLongNameAt(parentDirFd, cName)
 	}
-	// The now-deleted directory may have been in the DirIV cache. Clear it.
-	fs.nameTransform.DirIVCache.Clear()
 	return fuse.OK
 }
 
-// OpenDir implements pathfs.FileSystem
+// OpenDir - FUSE call
+//
+// This function is symlink-safe through use of openBackingDir() and
+// ReadDirIVAt().
 func (fs *FS) OpenDir(dirName string, context *fuse.Context) ([]fuse.DirEntry, fuse.Status) {
 	tlog.Debug.Printf("OpenDir(%s)", dirName)
-	cDirName, err := fs.encryptPath(dirName)
+	parentDirFd, cDirName, err := fs.openBackingDir(dirName)
 	if err != nil {
 		return nil, fuse.ToStatus(err)
 	}
+	defer syscall.Close(parentDirFd)
 	// Read ciphertext directory
-	cDirAbsPath := filepath.Join(fs.args.Cipherdir, cDirName)
 	var cipherEntries []fuse.DirEntry
 	var status fuse.Status
-	fd, err := syscall.Open(cDirAbsPath, syscall.O_RDONLY|syscall.O_NOFOLLOW, 0)
+	fd, err := syscallcompat.Openat(parentDirFd, cDirName, syscall.O_RDONLY|syscall.O_DIRECTORY|syscall.O_NOFOLLOW, 0)
 	if err != nil {
 		return nil, fuse.ToStatus(err)
 	}
@@ -279,23 +285,16 @@ func (fs *FS) OpenDir(dirName string, context *fuse.Context) ([]fuse.DirEntry, f
 	// Get DirIV (stays nil if PlaintextNames is used)
 	var cachedIV []byte
 	if !fs.args.PlaintextNames {
-		cachedIV, _ = fs.nameTransform.DirIVCache.Lookup(dirName)
-		if cachedIV == nil {
-			// Read the DirIV from disk and store it in the cache
-			fs.dirIVLock.RLock()
-			cachedIV, err = nametransform.ReadDirIV(cDirAbsPath)
-			if err != nil {
-				fs.dirIVLock.RUnlock()
-				// The directory itself does not exist
-				if err == syscall.ENOENT {
-					return nil, fuse.ENOENT
-				}
-				// Any other problem warrants an error message
-				tlog.Warn.Printf("OpenDir %q: could not read gocryptfs.diriv: %v", cDirName, err)
-				return nil, fuse.EIO
+		// Read the DirIV from disk
+		cachedIV, err = nametransform.ReadDirIVAt(fd)
+		if err != nil {
+			// The directory itself does not exist
+			if err == syscall.ENOENT {
+				return nil, fuse.ENOENT
 			}
-			fs.nameTransform.DirIVCache.Store(dirName, cachedIV, cDirName)
-			fs.dirIVLock.RUnlock()
+			// Any other problem warrants an error message
+			tlog.Warn.Printf("OpenDir %q: could not read gocryptfs.diriv: %v", cDirName, err)
+			return nil, fuse.EIO
 		}
 	}
 	// Decrypted directory entries
@@ -322,7 +321,7 @@ func (fs *FS) OpenDir(dirName string, context *fuse.Context) ([]fuse.DirEntry, f
 			isLong = nametransform.NameType(cName)
 		}
 		if isLong == nametransform.LongNameContent {
-			cNameLong, err := nametransform.ReadLongName(filepath.Join(cDirAbsPath, cName))
+			cNameLong, err := nametransform.ReadLongNameAt(fd, cName)
 			if err != nil {
 				tlog.Warn.Printf("OpenDir %q: invalid entry %q: Could not read .name: %v",
 					cDirName, cName, err)
