@@ -24,6 +24,8 @@ const (
 
 // HashLongName - take the hash of a long string "name" and return
 // "gocryptfs.longname.[sha256]"
+//
+// This function does not do any I/O.
 func (n *NameTransform) HashLongName(name string) string {
 	hashBin := sha256.Sum256([]byte(name))
 	hashBase64 := n.B64.EncodeToString(hashBin[:])
@@ -47,6 +49,8 @@ const (
 // gocryptfs.longname.[sha256]  ........ LongNameContent (content of a long name file)
 // gocryptfs.longname.[sha256].name .... LongNameFilename (full file name of a long name file)
 // else ................................ LongNameNone (normal file)
+//
+// This function does not do any I/O.
 func NameType(cName string) int {
 	if !strings.HasPrefix(cName, longNamePrefix) {
 		return LongNameNone
@@ -59,23 +63,32 @@ func NameType(cName string) int {
 
 // IsLongContent returns true if "cName" is the content store of a long name
 // file (looks like "gocryptfs.longname.[sha256]").
+//
+// This function does not do any I/O.
 func IsLongContent(cName string) bool {
 	return NameType(cName) == LongNameContent
 }
 
-// ReadLongName - read "$path.name"
-func ReadLongName(path string) (string, error) {
-	path += LongNameSuffix
-	fd, err := os.Open(path)
-	if err != nil {
-		return "", err
+// ReadLongName - read cName + ".name" from the directory opened as dirfd.
+//
+// Symlink-safe through Openat().
+func ReadLongNameAt(dirfd int, cName string) (string, error) {
+	cName += LongNameSuffix
+	var f *os.File
+	{
+		fd, err := syscallcompat.Openat(dirfd, cName, syscall.O_RDONLY|syscall.O_NOFOLLOW, 0)
+		if err != nil {
+			return "", err
+		}
+		f = os.NewFile(uintptr(fd), "")
+		// fd runs out of scope here
 	}
-	defer fd.Close()
+	defer f.Close()
 	// 256 (=255 padded to 16) bytes base64-encoded take 344 bytes: "AAAAAAA...AAA=="
 	lim := 344
 	// Allocate a bigger buffer so we see whether the file is too big
 	buf := make([]byte, lim+1)
-	n, err := fd.ReadAt(buf, 0)
+	n, err := f.ReadAt(buf, 0)
 	if err != nil && err != io.EOF {
 		return "", err
 	}
@@ -88,9 +101,11 @@ func ReadLongName(path string) (string, error) {
 	return string(buf[0:n]), nil
 }
 
-// DeleteLongName deletes "hashName.name".
-func DeleteLongName(dirfd *os.File, hashName string) error {
-	err := syscallcompat.Unlinkat(int(dirfd.Fd()), hashName+LongNameSuffix, 0)
+// DeleteLongName deletes "hashName.name" in the directory opened at "dirfd".
+//
+// This function is symlink-safe through the use of Unlinkat().
+func DeleteLongNameAt(dirfd int, hashName string) error {
+	err := syscallcompat.Unlinkat(dirfd, hashName+LongNameSuffix, 0)
 	if err != nil {
 		tlog.Warn.Printf("DeleteLongName: %v", err)
 	}
@@ -99,8 +114,10 @@ func DeleteLongName(dirfd *os.File, hashName string) error {
 
 // WriteLongName encrypts plainName and writes it into "hashName.name".
 // For the convenience of the caller, plainName may also be a path and will be
-// converted internally.
-func (n *NameTransform) WriteLongName(dirfd *os.File, hashName string, plainName string) (err error) {
+// Base()named internally.
+//
+// This function is symlink-safe through the use of Openat().
+func (n *NameTransform) WriteLongNameAt(dirfd int, hashName string, plainName string) (err error) {
 	plainName = filepath.Base(plainName)
 
 	// Encrypt the basename
@@ -111,8 +128,8 @@ func (n *NameTransform) WriteLongName(dirfd *os.File, hashName string, plainName
 	cName := n.EncryptName(plainName, dirIV)
 
 	// Write the encrypted name into hashName.name
-	fdRaw, err := syscallcompat.Openat(int(dirfd.Fd()), hashName+LongNameSuffix,
-		syscall.O_WRONLY|syscall.O_CREAT|syscall.O_EXCL, 0600)
+	fdRaw, err := syscallcompat.Openat(dirfd, hashName+LongNameSuffix,
+		syscall.O_WRONLY|syscall.O_CREAT|syscall.O_EXCL, 0400)
 	if err != nil {
 		// Don't warn if the file already exists - this is allowed for renames
 		// and should be handled by the caller.
@@ -122,10 +139,20 @@ func (n *NameTransform) WriteLongName(dirfd *os.File, hashName string, plainName
 		return err
 	}
 	fd := os.NewFile(uintptr(fdRaw), hashName+LongNameSuffix)
-	defer fd.Close()
 	_, err = fd.Write([]byte(cName))
 	if err != nil {
+		fd.Close()
 		tlog.Warn.Printf("WriteLongName: Write: %v", err)
+		// Delete incomplete longname file
+		syscallcompat.Unlinkat(dirfd, hashName+LongNameSuffix, 0)
+		return err
 	}
-	return err
+	err = fd.Close()
+	if err != nil {
+		tlog.Warn.Printf("WriteLongName: Close: %v", err)
+		// Delete incomplete longname file
+		syscallcompat.Unlinkat(dirfd, hashName+LongNameSuffix, 0)
+		return err
+	}
+	return nil
 }
