@@ -9,16 +9,19 @@ import (
 
 	"github.com/rfjakob/eme"
 
-	"github.com/rfjakob/gocryptfs/internal/nametransform/dirivcache"
 	"github.com/rfjakob/gocryptfs/internal/tlog"
 )
 
+const (
+	// Like ext4, we allow at most 255 bytes for a file name.
+	NameMax = 255
+)
+
 // NameTransform is used to transform filenames.
 type NameTransform struct {
-	emeCipher  *eme.EMECipher
-	longNames  bool
-	DirIVCache dirivcache.DirIVCache
-	// B64 = either base64.URLEncoding or base64.RawURLEncoding, depeding
+	emeCipher *eme.EMECipher
+	longNames bool
+	// B64 = either base64.URLEncoding or base64.RawURLEncoding, depending
 	// on the Raw64 feature flag
 	B64 *base64.Encoding
 }
@@ -74,11 +77,10 @@ func (n *NameTransform) DecryptName(cipherName string, iv []byte) (string, error
 }
 
 // EncryptName encrypts "plainName", returns a base64-encoded "cipherName64".
-// Used internally by EncryptPathDirIV().
 // The encryption is either CBC or EME, depending on "useEME".
 //
-// This function is exported because fusefrontend needs access to the full (not hashed)
-// name if longname is used. Otherwise you should use EncryptPathDirIV()
+// This function is exported because in some cases, fusefrontend needs access
+// to the full (not hashed) name if longname is used.
 func (n *NameTransform) EncryptName(plainName string, iv []byte) (cipherName64 string) {
 	bin := []byte(plainName)
 	bin = pad16(bin)
