@@ -21,9 +21,9 @@ var testPw = []byte("test")
 
 func TestMain(m *testing.M) {
 	test_helpers.ResetTmpDir(false)
-	before := test_helpers.ListFds()
+	before := test_helpers.ListFds(0)
 	r := m.Run()
-	after := test_helpers.ListFds()
+	after := test_helpers.ListFds(0)
 	if len(before) != len(after) {
 		fmt.Printf("fd leak? before, after:\n%v\n%v\n", before, after)
 		os.Exit(1)
@@ -34,7 +34,7 @@ func TestMain(m *testing.M) {
 // Test -init flag
 func TestInit(t *testing.T) {
 	dir := test_helpers.InitFS(t)
-	_, c, err := configfile.Load(dir+"/"+configfile.ConfDefaultName, testPw)
+	_, c, err := configfile.LoadAndDecrypt(dir+"/"+configfile.ConfDefaultName, testPw)
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -51,7 +51,7 @@ func TestInitDevRandom(t *testing.T) {
 // Test -init with -aessiv
 func TestInitAessiv(t *testing.T) {
 	dir := test_helpers.InitFS(t, "-aessiv")
-	_, c, err := configfile.Load(dir+"/"+configfile.ConfDefaultName, testPw)
+	_, c, err := configfile.LoadAndDecrypt(dir+"/"+configfile.ConfDefaultName, testPw)
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -63,7 +63,7 @@ func TestInitAessiv(t *testing.T) {
 // Test -init with -reverse
 func TestInitReverse(t *testing.T) {
 	dir := test_helpers.InitFS(t, "-reverse")
-	_, c, err := configfile.Load(dir+"/"+configfile.ConfReverseName, testPw)
+	_, c, err := configfile.LoadAndDecrypt(dir+"/"+configfile.ConfReverseName, testPw)
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -495,3 +495,71 @@ func TestExcludeForward(t *testing.T) {
 	}
 	t.Log(err)
 }
+
+// Check that the config file can be read from a named pipe.
+// Make sure bug https://github.com/rfjakob/gocryptfs/issues/258 does not come
+// back.
+func TestConfigPipe(t *testing.T) {
+	dir := test_helpers.InitFS(t)
+	mnt := dir + ".mnt"
+	err := os.Mkdir(mnt, 0700)
+	if err != nil {
+		t.Fatal(err)
+	}
+	bashLine := fmt.Sprintf("%s -q -extpass \"echo test\" -config <(cat %s/gocryptfs.conf) %s %s", test_helpers.GocryptfsBinary, dir, dir, mnt)
+	cmd := exec.Command("bash", "-c", bashLine)
+	cmd.Stdout = os.Stdout
+	cmd.Stderr = os.Stdout
+	err = cmd.Run()
+	exitCode := test_helpers.ExtractCmdExitCode(err)
+	if exitCode != 0 {
+		t.Errorf("bash command\n%q\nresulted in exit code %d", bashLine, exitCode)
+		return
+	}
+	test_helpers.UnmountPanic(mnt)
+}
+
+// Ciphertext dir and mountpoint contains a comma
+// https://github.com/rfjakob/gocryptfs/issues/262
+func TestComma(t *testing.T) {
+	dir0 := test_helpers.InitFS(t)
+	dir := dir0 + ",foo,bar"
+	err := os.Rename(dir0, dir)
+	if err != nil {
+		t.Fatal(err)
+	}
+	mnt := dir + ".mnt"
+	err = test_helpers.Mount(dir, mnt, false, "-extpass", "echo test", "-wpanic=0")
+	if err != nil {
+		t.Fatalf("Failed to mount %q on %q: %v", dir, mnt, err)
+	}
+	test_helpers.UnmountPanic(mnt)
+}
+
+// Mount with idle timeout 10ms and check that the process exits by itself
+// within 5 seconds.
+func TestIdle(t *testing.T) {
+	dir := test_helpers.InitFS(t)
+	mnt := dir + ".mnt"
+	err := os.Mkdir(mnt, 0700)
+	if err != nil {
+		t.Fatal(err)
+	}
+	cmd := exec.Command(test_helpers.GocryptfsBinary,
+		"-q", "-nosyslog", "-fg", "-extpass", "echo test", "-i", "10ms", dir, mnt)
+	cmd.Stdout = os.Stdout
+	cmd.Stderr = os.Stderr
+	err = cmd.Start()
+	if err != nil {
+		t.Fatal(err)
+	}
+	timer := time.AfterFunc(5*time.Second, func() {
+		t.Error("timeout waiting for umount")
+		cmd.Process.Kill()
+	})
+	err = cmd.Wait()
+	timer.Stop()
+	if err != nil {
+		t.Error(err)
+	}
+}
