@@ -23,8 +23,9 @@ import (
 	"syscall"
 	"testing"
 
+	"golang.org/x/sys/unix"
+
 	"github.com/rfjakob/gocryptfs/internal/stupidgcm"
-	"github.com/rfjakob/gocryptfs/internal/syscallcompat"
 	"github.com/rfjakob/gocryptfs/tests/test_helpers"
 )
 
@@ -72,9 +73,11 @@ func TestMain(m *testing.M) {
 		opts = append(opts, fmt.Sprintf("-aessiv=%v", testcase.aessiv))
 		opts = append(opts, fmt.Sprintf("-raw64=%v", testcase.raw64))
 		test_helpers.MountOrExit(test_helpers.DefaultCipherDir, test_helpers.DefaultPlainDir, opts...)
-		before := test_helpers.ListFds()
+		before := test_helpers.ListFds(0)
 		r := m.Run()
-		after := test_helpers.ListFds()
+		// Catch fd leaks in the tests. NOTE: this does NOT catch leaks in
+		// the gocryptfs FUSE process, but only in the tests that access it!
+		after := test_helpers.ListFds(0)
 		if len(before) != len(after) {
 			fmt.Printf("fd leak? before, after:\n%v\n%v\n", before, after)
 			os.Exit(1)
@@ -95,13 +98,13 @@ func testWriteN(t *testing.T, fn string, n int) string {
 	}
 
 	d := make([]byte, n)
-	written, err := file.Write(d)
-	if err != nil || written != len(d) {
-		t.Errorf("err=\"%s\", written=%d", err, written)
+	_, err = file.Write(d)
+	if err != nil {
+		t.Fatal(err)
 	}
 	err = file.Close()
 	if err != nil {
-		t.Error(err)
+		t.Fatal(err)
 	}
 
 	test_helpers.VerifySize(t, test_helpers.DefaultPlainDir+"/"+fn, n)
@@ -204,151 +207,6 @@ func TestTruncate(t *testing.T) {
 	}
 }
 
-const (
-	// From man statfs
-	TMPFS_MAGIC      = 0x01021994
-	EXT4_SUPER_MAGIC = 0xef53
-)
-
-// isWellKnownFS decides if the backing filesystem is well-known.
-// The expected allocated sizes are only valid on tmpfs and ext4. btrfs
-// gives different results, but that's not an error.
-func isWellKnownFS(fn string) bool {
-	var fs syscall.Statfs_t
-	err := syscall.Statfs(fn, &fs)
-	if err != nil {
-		panic(err)
-	}
-	if fs.Type == EXT4_SUPER_MAGIC || fs.Type == TMPFS_MAGIC {
-		return true
-	}
-	return false
-}
-
-const FALLOC_DEFAULT = 0x00
-const FALLOC_FL_KEEP_SIZE = 0x01
-
-func TestFallocate(t *testing.T) {
-	if runtime.GOOS == "darwin" {
-		t.Skipf("OSX does not support fallocate")
-	}
-	fn := test_helpers.DefaultPlainDir + "/fallocate"
-	file, err := os.Create(fn)
-	if err != nil {
-		t.FailNow()
-	}
-	defer file.Close()
-	wellKnown := isWellKnownFS(test_helpers.DefaultCipherDir)
-	fd := int(file.Fd())
-	nBytes := test_helpers.Du(t, fd)
-	if nBytes != 0 {
-		t.Fatalf("Empty file has %d bytes", nBytes)
-	}
-	// Allocate 30 bytes, keep size
-	// gocryptfs ||        (0 blocks)
-	//      ext4 |  d   |  (1 block)
-	err = syscallcompat.Fallocate(fd, FALLOC_FL_KEEP_SIZE, 0, 30)
-	if err != nil {
-		t.Error(err)
-	}
-	var want int64
-	nBytes = test_helpers.Du(t, fd)
-	want = 4096
-	if nBytes != want {
-		t.Errorf("Expected %d allocated bytes, have %d", want, nBytes)
-	}
-	test_helpers.VerifySize(t, fn, 0)
-	// Three ciphertext blocks. The middle one should be a file hole.
-	// gocryptfs |  h   |   h  | d|   (1 block)
-	//      ext4 |  d  |  h  |  d  |  (2 blocks)
-	// (Note that gocryptfs blocks are slightly bigger than the ext4 blocks,
-	// but the last one is partial)
-	err = file.Truncate(9000)
-	if err != nil {
-		t.Fatal(err)
-	}
-	nBytes = test_helpers.Du(t, fd)
-	want = 2 * 4096
-	if wellKnown && nBytes != want {
-		t.Errorf("Expected %d allocated bytes, have %d", want, nBytes)
-	}
-	if md5 := test_helpers.Md5fn(fn); md5 != "5420afa22f6423a9f59e669540656bb4" {
-		t.Errorf("Wrong md5 %s", md5)
-	}
-	// Allocate the whole file space
-	// gocryptfs |  h   |   h  | d|   (1 block)
-	//      ext4 |  d  |  d  |  d  |  (3 blocks
-	err = syscallcompat.Fallocate(fd, FALLOC_DEFAULT, 0, 9000)
-	if err != nil {
-		t.Fatal(err)
-	}
-	nBytes = test_helpers.Du(t, fd)
-	want = 3 * 4096
-	if nBytes != want {
-		t.Errorf("Expected %d allocated bytes, have %d", want, nBytes)
-	}
-	// Neither apparent size nor content should have changed
-	test_helpers.VerifySize(t, fn, 9000)
-	if md5 := test_helpers.Md5fn(fn); md5 != "5420afa22f6423a9f59e669540656bb4" {
-		t.Errorf("Wrong md5 %s", md5)
-	}
-
-	// Partial block on the end. The first ext4 block is dirtied by the header.
-	// gocryptfs |  h   |   h  | d|   (1 block)
-	//      ext4 |  d  |  h  |  d  |  (2 blocks)
-	file.Truncate(0)
-	file.Truncate(9000)
-	nBytes = test_helpers.Du(t, fd)
-	want = 2 * 4096
-	if wellKnown && nBytes != want {
-		t.Errorf("Expected %d allocated bytes, have %d", want, nBytes)
-	}
-	// Allocate 10 bytes in the second block
-	// gocryptfs |  h   |   h  | d|   (1 block)
-	//      ext4 |  d  |  d  |  d  |  (3 blocks)
-	syscallcompat.Fallocate(fd, FALLOC_DEFAULT, 5000, 10)
-	nBytes = test_helpers.Du(t, fd)
-	want = 3 * 4096
-	if wellKnown && nBytes != want {
-		t.Errorf("Expected %d allocated bytes, have %d", want, nBytes)
-	}
-	// Neither apparent size nor content should have changed
-	test_helpers.VerifySize(t, fn, 9000)
-	if md5 := test_helpers.Md5fn(fn); md5 != "5420afa22f6423a9f59e669540656bb4" {
-		t.Errorf("Wrong md5 %s", md5)
-	}
-	// Grow the file to 4 blocks
-	// gocryptfs |  h   |  h   |  d   |d|  (2 blocks)
-	//      ext4 |  d  |  d  |  d  |  d  | (4 blocks)
-	syscallcompat.Fallocate(fd, FALLOC_DEFAULT, 15000, 10)
-	nBytes = test_helpers.Du(t, fd)
-	want = 4 * 4096
-	if wellKnown && nBytes != want {
-		t.Errorf("Expected %d allocated bytes, have %d", want, nBytes)
-	}
-	test_helpers.VerifySize(t, fn, 15010)
-	if md5 := test_helpers.Md5fn(fn); md5 != "c4c44c7a41ab7798a79d093eb44f99fc" {
-		t.Errorf("Wrong md5 %s", md5)
-	}
-	// Shrinking a file using fallocate should have no effect
-	for _, off := range []int64{0, 10, 2000, 5000} {
-		for _, sz := range []int64{0, 1, 42, 6000} {
-			syscallcompat.Fallocate(fd, FALLOC_DEFAULT, off, sz)
-			test_helpers.VerifySize(t, fn, 15010)
-			if md5 := test_helpers.Md5fn(fn); md5 != "c4c44c7a41ab7798a79d093eb44f99fc" {
-				t.Errorf("Wrong md5 %s", md5)
-			}
-		}
-	}
-	// Cleanup
-	syscall.Unlink(fn)
-	if !wellKnown {
-		// Even though most tests have been executed still, inform the user
-		// that some were disabled
-		t.Skipf("backing fs is not ext4 or tmpfs, skipped some disk-usage checks\n")
-	}
-}
-
 func TestAppend(t *testing.T) {
 	fn := test_helpers.DefaultPlainDir + "/append"
 	file, err := os.Create(fn)
@@ -514,31 +372,59 @@ func TestFilenameEncryption(t *testing.T) {
 	}
 }
 
-// Test Mkdir and Rmdir
-func testMkdirRmdir(t *testing.T) {
-	test_helpers.TestMkdirRmdir(t, test_helpers.DefaultPlainDir)
-}
-
 // Test Rename
-func testRename(t *testing.T) {
+func TestRename(t *testing.T) {
 	test_helpers.TestRename(t, test_helpers.DefaultPlainDir)
 }
 
-// Overwrite an empty directory with another directory
-func TestDirOverwrite(t *testing.T) {
-	dir1 := test_helpers.DefaultPlainDir + "/DirOverwrite1"
-	dir2 := test_helpers.DefaultPlainDir + "/DirOverwrite2"
-	err := os.Mkdir(dir1, 0777)
+// Test that names of all lengths work
+func TestNameLengths(t *testing.T) {
+	f, err := os.Open(test_helpers.DefaultPlainDir)
 	if err != nil {
 		t.Fatal(err)
 	}
-	err = os.Mkdir(dir2, 0777)
+	entries, err := f.Readdirnames(0)
 	if err != nil {
 		t.Fatal(err)
 	}
-	err = syscall.Rename(dir1, dir2)
-	if err != nil {
-		t.Fatal(err)
+	f.Close()
+	cnt1 := len(entries)
+
+	wd := test_helpers.DefaultPlainDir + "/"
+	name := "x"
+	for len(name) < 2000 {
+		f, err := os.Create(wd + name + "x")
+		if err != nil {
+			break
+		}
+		name = name + "x"
+		f.Close()
+		f, err = os.Open(test_helpers.DefaultPlainDir)
+		if err != nil {
+			t.Fatal(err)
+		}
+		// In v1.7-rc2, we had a bug that allowed creation of too-long names.
+		// This threw errors in like this in READDIR:
+		//
+		//   OpenDir ".": invalid entry "gocryptfs.longname.wrE-izsR9ciEkP7JSCFDrk_d_Nj4mQo1dGY6hjuixAU=":
+		//   Could not read .name: ReadLongName: size=345 > limit=344
+		//
+		entries, err = f.Readdirnames(0)
+		if err != nil {
+			t.Fatal(err)
+		}
+		f.Close()
+		cnt2 := len(entries)
+		if cnt2 != cnt1+1 {
+			t.Fatalf("len=%d: expected %d dir entries, have %d: %v", len(name), cnt1+1, cnt2, entries)
+		}
+		err = syscall.Unlink(wd + name)
+		if err != nil {
+			t.Fatal(err)
+		}
+	}
+	if len(name) != 255 {
+		t.Errorf("maxlen=%d", len(name))
 	}
 }
 
@@ -699,7 +585,7 @@ func TestLchown(t *testing.T) {
 }
 
 // Set nanoseconds by path, symlink
-func Symlink(t *testing.T) {
+func TestUtimesNanoSymlink(t *testing.T) {
 	if runtime.GOOS == "darwin" {
 		t.Skipf("MacOS \"touch\" does not support \"--no-dereference\"")
 	}
@@ -767,7 +653,7 @@ func doTestUtimesNano(t *testing.T, path string) {
 	for i, tc := range utimeTestcases {
 		err := syscall.UtimesNano(path, tc.in[:])
 		if err != nil {
-			t.Fatal(err)
+			t.Fatalf("%q: %v", path, err)
 		}
 		var st syscall.Stat_t
 		err = syscall.Stat(path, &st)
@@ -883,3 +769,60 @@ func TestMagicNames(t *testing.T) {
 		}
 	}
 }
+
+// Test that chmod works correctly
+func TestChmod(t *testing.T) {
+	// Note: t.Name() is not available before in Go 1.8
+	tName := "TestChmod"
+	path := test_helpers.DefaultPlainDir + "/" + tName
+	file, err := os.Create(path)
+	if err != nil {
+		t.Fatal(err)
+	}
+	file.Close()
+	modes := []os.FileMode{0777, 0707, 0606, 0666, 0444, 0000, 0111, 0123, 0321}
+	for _, modeWant := range modes {
+		fi, err := os.Stat(path)
+		if err != nil {
+			t.Fatal(err)
+		}
+		err = syscall.Chmod(path, uint32(modeWant))
+		if err != nil {
+			t.Errorf("chmod %03o -> %03o failed: %v", fi.Mode(), modeWant, err)
+			continue
+		}
+		fi, err = os.Stat(path)
+		if err != nil {
+			t.Fatal(err)
+		}
+		modeHave := fi.Mode()
+		if modeHave != modeWant {
+			t.Errorf("modeHave %#o != modeWant %#o", modeHave, modeWant)
+		}
+	}
+}
+
+// Test that access(2) works correctly
+func TestAccess(t *testing.T) {
+	// Note: t.Name() is not available before in Go 1.8
+	tName := "TestAccess"
+	path := test_helpers.DefaultPlainDir + "/" + tName
+	file, err := os.Create(path)
+	if err != nil {
+		t.Fatal(err)
+	}
+	defer file.Close()
+
+	err = unix.Access(path, unix.F_OK)
+	if err != nil {
+		t.Error(err)
+	}
+	err = unix.Access(path, unix.R_OK)
+	if err != nil {
+		t.Error(err)
+	}
+	err = unix.Access(path, unix.X_OK)
+	if err == nil {
+		t.Error("X_OK should have failed")
+	}
+}
